PROGRAM uudecode;

USES
  Crt;

  CONST defaultSuffix = '.uue';
        offset = 32;

  TYPE STRING80 = STRING[80];

  VAR infile: Text;
      outfile: FILE OF Byte;
      lineNum: Integer;
      Line: STRING80;

  PROCEDURE Abort(message: STRING80);

    BEGIN {abort}
      Writeln;
      IF lineNum > 0 THEN Write('Line ', lineNum, ': ');
      Writeln(message);
      Halt
    END; {Abort}

  PROCEDURE NextLine(VAR s: STRING80);

    BEGIN {NextLine}
      LineNum := Succ(LineNum);
      Write('.');
      Readln(infile, s)
    END; {NextLine}

  PROCEDURE Init;

    PROCEDURE GetInFile;

      VAR infilename: STRING80;

      BEGIN {GetInFile}
        IF ParamCount = 0 THEN abort ('Usage: uudecode <filename>');
        infilename := ParamStr(1);
        IF Pos('.', infilename) = 0
          THEN infilename := Concat(infilename, defaultSuffix);
        Assign(infile, infilename);
        {$i-}
        Reset(infile);
        {$i+}
        IF IOresult > 0 THEN abort (Concat('Can''t open ', infilename));
        Writeln ('Decoding ', infilename)
      END; {GetInFile}

    PROCEDURE GetOutFile;

      VAR header, mode, outfilename: STRING80;
          ch: Char;

      PROCEDURE ParseHeader;

        VAR index: Integer;

        PROCEDURE NextWord(VAR Word:STRING80; VAR index: Integer);

          BEGIN {nextword}
            Word := '';
            WHILE header[index] = ' ' DO
              BEGIN
                index := Succ(index);
                IF index > Length(header) THEN abort ('Incomplete header')
              END;
            WHILE header[index] <> ' ' DO
              BEGIN
                Word := Concat(Word, header[index]);
                index := Succ(index)
              END
          END; {NextWord}

        BEGIN {ParseHeader}
          header := Concat(header, ' ');
          index := 7;
          NextWord(mode, index);
          NextWord(outfilename, index)
        END; {ParseHeader}

      BEGIN {GetOutFile}
        IF EOF(infile) THEN abort('Nothing to decode.');
        NextLine (header);
        WHILE NOT ((Copy(header, 1, 6) = 'begin ') OR EOF(infile)) DO
          NextLine(header);
        Writeln;
        IF EOF(infile) THEN abort('Nothing to decode.');
        ParseHeader;
        Assign(outfile, outfilename);
        Writeln ('Destination is ', outfilename);
        {$i-}
        Reset(outfile);
        {$i+}
        IF IOresult = 0 THEN
          BEGIN
            Write ('Overwrite current ', outfilename, '? [Y/N] ');
            REPEAT
              ch:=ReadKey;
              ch := UpCase(ch)
            UNTIL ch IN ['Y', 'N'];
            Writeln(ch);
            IF ch = 'N' THEN abort ('Overwrite cancelled.')
          END;
        Rewrite (outfile);
      END; {GetOutFile}

    BEGIN {init}
      lineNum := 0;
      GetInFile;
      GetOutFile;
    END; { init}

  FUNCTION CheckLine: Boolean;

    BEGIN {CheckLine}
      IF Line = '' THEN abort ('Blank line in file');
      CheckLine := NOT (Line[1] IN [' ', '`'])
    END; {CheckLine}


  PROCEDURE DecodeLine;

    VAR lineIndex, byteNum, count, i: Integer;
        chars: ARRAY [0..3] OF Byte;
        hunk: ARRAY [0..2] OF Word;

{    procedure debug;

      var i: integer;

      procedure writebin(x: byte);

        var i: integer;

        begin
          for i := 1 to 8 do
            begin
              write ((x and $80) shr 7);
              x := x shl 1
            end;
          write (' ')
        end;

      begin
        writeln;
        for i := 0 to 3 do writebin(chars[i]);
        writeln;
        for i := 0 to 2 do writebin(hunk[i]);
        writeln
      end;      }

    FUNCTION nextch: Char;

      BEGIN {nextch}
        lineIndex := Succ(lineIndex);
        IF lineIndex > Length(Line) THEN abort('Line too short.');
        IF NOT (Line[lineindex] IN [' '..'`'])
          THEN abort('Illegal character in line.');
{        write(line[lineindex]:2);}
        IF Line[lineindex] = '`' THEN nextch := ' '
                                  ELSE nextch := Line[lineIndex]
      END; {nextch}

    PROCEDURE DecodeByte;

      PROCEDURE GetNextHunk;

        VAR i: Integer;

        BEGIN {GetNextHunk}
          FOR i := 0 TO 3 DO chars[i] := Ord(nextch) - offset;
          hunk[0] := (chars[0] SHL 2) + (chars[1] SHR 4);
          hunk[1] := (chars[1] SHL 4) + (chars[2] SHR 2);
          hunk[2] := (chars[2] SHL 6) + chars[3];
          byteNum := 0  {;
          debug          }
        END; {GetNextHunk}

      BEGIN {DecodeByte}
        IF byteNum = 3 THEN GetNextHunk;
        Write (outfile, Byte(hunk[byteNum]));
        {writeln(bytenum, ' ', hunk[byteNum]);}
        byteNum := Succ(byteNum)
      END; {DecodeByte}

    BEGIN {DecodeLine}
      lineIndex := 0;
      byteNum := 3;
      count := (Ord(nextch) - offset);
      FOR i := 1 TO count DO DecodeByte
    END; {DecodeLine}

  PROCEDURE terminate;

    VAR trailer: STRING80;

    BEGIN {terminate}
      IF EOF(infile) THEN abort ('Abnormal end.');
      NextLine (trailer);
      IF Length (trailer) < 3 THEN abort ('Abnormal end.');
      IF Copy (trailer, 1, 3) <> 'end' THEN abort ('Abnormal end.');
      Close (infile);
      Close (outfile)
    END;

  BEGIN {uudecode}
    init;
    NextLine(Line);
    WHILE CheckLine DO
      BEGIN
        DecodeLine;
        NextLine(Line)
      END;
    terminate
  END.
